from distutils.core import setup

setup(
  name = 'mitochondria',
  packages = ['mitochondria'],
  version = '0.0.5',
  description = '',
  author = 'Liling Tan',
  license = 'MIT',
  url = 'https://github.com/alvations/mitochondria',
  download_url = '',
  keywords = [],
  classifiers = [],
)
